/**
  ****************************************************************************************************
  * @file    tsi.c
  * @author  FMSH Application Team
  * @brief   Source file of FL Driver Library
  ****************************************************************************************************
  * @attention
  *
  * Copyright (c) [2021] [Fudan Microelectronics]
  * THIS SOFTWARE is licensed under Mulan PSL v2.
  * You can use this software according to the terms and conditions of the Mulan PSL v2.
  * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
  * EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
  * MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
  * See the Mulan PSL v2 for more details.
  *
  ****************************************************************************************************
  */
/* Includes ----------------------------------------------------------------------------------------*/
#include "tsi.h"
#include "tsi_objects.h"
#include "tsi_widget.h"
#include "tsi_calibration.h"

#include <string.h>

/* Macros ---------------------------------------------------------------------------------------------*/

/* Defines --------------------------------------------------------------------------------------------*/

/* Exported Functions ---------------------------------------------------------------------------------*/
TSI_RetCodeType TSI_Init(void)
{
    int i;

    /* Init clock & pins on first entry -------------------------------------*/
    if (TSI_RDSTAT_LIB_STAT() == TSI_LIB_RESET)
    {
        /* Init module clock */
        FL_TSI_ClockInit(TSI_Module.hardwareCfg.opClkSource,
                         TSI_Module.hardwareCfg.opClkPrescaler,
                         TSI_Module.hardwareCfg.opClkFreq);

        /* Init used pins */
        for (i = 0; i < TSI_TOTAL_PINS_NUM; i++)
        {
            FL_TSI_PinInit(TSI_Pins[i].pinUniqueId);
        }

        /* Init NVIC */
        FL_TSI_InterruptInit(TSI_INTERRUPT_PRIORITY);
    }

    /* Module reset ---------------------------------------------------------*/
    FL_TSI_ModuleReset();

    /* Init objects ---------------------------------------------------------*/
    memcpy(&TSI_Module, &TSI_ModuleConstInit, sizeof(TSI_ModuleTypeDef));
    memcpy(&TSI_WidgetList, &TSI_WidgetListConstInit, sizeof(TSI_WidgetListTypeDef));
    memcpy(&TSI_SensorList, &TSI_SensorListConstInit, sizeof(TSI_SensorListTypeDef));

    /* Configure hardware ---------------------------------------------------*/
    /* Set sample clock source */
#if (TSI_SAMPLE_CLK_SOURCE == TSI_SAMPLE_CLK_DIRECT)
    FL_TSI_SetSampleClockSource(TSI, FL_TSI_SAMPLE_CLK_SOURCE_FIXED_CLK);
#else
    FL_TSI_SetSampleClockSource(TSI, FL_TSI_SAMPLE_CLK_SOURCE_PRS_CLK);
#endif
    
#if (TSI_USE_SHIELD)
    /* Enable shield driver output */
    FL_TSI_EnableShield(TSI);
#endif
    
    /* Set inactive sensor connection */
    FL_TSI_SetInactiveSensorConnection(TSI, FL_TSI_INACTIVE_SENSOR_CONNECTION_GROUND);

    /* Continuous scan */
    FL_TSI_SetScanMode(TSI, FL_TSI_SCAN_MODE_CONTINUOUS);

    /* Stop immediately */
    FL_TSI_SetStopMode(TSI, FL_TSI_STOP_MODE_IMMEDIATELY);
    
    /* Scan timing */
    FL_TSI_WriteScanIdleTime(TSI, TSI_Module.scanCfg.sampleIdleTime);
    FL_TSI_WriteScanIntervalTime(TSI, TSI_Module.scanCfg.sampleInterval);

    /* Init TSI library -----------------------------------------------------*/
    /* Set lib status */
    TSI_WRSTAT_LIB_STAT(TSI_LIB_INIT);

#if (TSI_CALIB_METHOD == TSI_CALIB_METHOD_HARDWARE)
    
    /* Hardware parameter auto calibration */
    TSI_CalibrateAllWidgets();

#endif

    /* recalculate and set scan period */
    TSI_AutoSetScanPeriod();
    if(FL_TSI_ReadScanPeriod(TSI) < TSI_Module.scanCfg.scanPeriod)
    {
        FL_TSI_WriteScanPeriod(TSI, TSI_Module.scanCfg.scanPeriod);
    }

    /* Reconfigure TSI, which will apply hw params and peform an initial scan */
    TSI_Widget_EnableAll();
    TSI_Reconfig();
    
    /* Disable all widget */
    TSI_Widget_DisableAll();

    /* Enable TSI */
    FL_TSI_Enable(TSI);

    return TSI_PASS;
}

TSI_RetCodeType TSI_Start(void)
{
    /* Set lib status */
    if (TSI_RDSTAT_LIB_STAT() != TSI_LIB_INIT && TSI_RDSTAT_LIB_STAT() != TSI_LIB_SUSPEND)
    {
        /* State machine corrupted */
        return TSI_FSM_CORRUPTED;
    }
    TSI_WRSTAT_LIB_STAT(TSI_LIB_RUNNING);

    /* Clear end of scan flag */
    TSI_CLRSTAT_SCAN_CPLT();

    /* Enable EOC & EOS interrupt */
    FL_TSI_ClearFlag_EndOfChannel(TSI);
    FL_TSI_EnableIT_EndOfChannel(TSI);
    FL_TSI_ClearFlag_EndOfSequence(TSI);
    FL_TSI_EnableIT_EndOfSequence(TSI);

    /* Start scanning */
    FL_TSI_Enable(TSI);
    FL_TSI_EnableScan(TSI);

    return TSI_PASS;
}

TSI_RetCodeType TSI_Suspend(void)
{
    /* Set lib status */
    if (TSI_RDSTAT_LIB_STAT() != TSI_LIB_RUNNING)
    {
        /* State machine corrupted */
        return TSI_FSM_CORRUPTED;
    }
    TSI_WRSTAT_LIB_STAT(TSI_LIB_SUSPEND);

    /* Stop scanning immediately */
    FL_TSI_SetStopMode(TSI, FL_TSI_STOP_MODE_IMMEDIATELY);
    FL_TSI_Disable(TSI);

    /* Disable EOC & EOS interrupt */
    FL_TSI_DisableIT_EndOfChannel(TSI);
    FL_TSI_ClearFlag_EndOfChannel(TSI);
    FL_TSI_DisableIT_EndOfSequence(TSI);
    FL_TSI_ClearFlag_EndOfSequence(TSI);

    /* Clear end of scan flag */
    TSI_CLRSTAT_SCAN_CPLT();

    return TSI_PASS;
}

TSI_RetCodeType TSI_Resume(void)
{
    return TSI_Start();
}

TSI_RetCodeType TSI_Reconfig(void)
{
    uint32_t timeout;
    uint32_t oldLibStat = TSI_RDSTAT_LIB_STAT();
    
    if (oldLibStat == TSI_LIB_RESET)
    {
        /* State machine corrupted */
        return TSI_FSM_CORRUPTED;
    }
    
    /* Suspend TSI */
    if (oldLibStat == TSI_LIB_RUNNING)
    {
        (void) TSI_Suspend();
    }
    
    /* Apply hardware configurations */
    TSI_Widget_ApplyConfigToAll();
    
    /* Set to single scan */
    FL_TSI_SetScanMode(TSI, FL_TSI_SCAN_MODE_SINGLE);

    /* Enable EOC & EOS interrupt */
    FL_TSI_ClearFlag_EndOfChannel(TSI);
    FL_TSI_EnableIT_EndOfChannel(TSI);
    FL_TSI_ClearFlag_EndOfSequence(TSI);
    FL_TSI_EnableIT_EndOfSequence(TSI);

    /* Start scanning */
    FL_TSI_Enable(TSI);
    FL_TSI_EnableScan(TSI);

    /* Wait the scanning to complete */
    timeout = TSI_STARTUP_TIMEOUT_VALUE;
    while (!TSI_GETSTAT_SCAN_CPLT() && timeout > 0)
    {
        timeout--;
    }

    if (timeout == 0)
    {
        return TSI_TIMEOUT;
    }

    /* Clear scan complete flag */
    TSI_CLRSTAT_SCAN_CPLT();

    /* Disable EOC & EOS interrupt */
    FL_TSI_DisableIT_EndOfChannel(TSI);
    FL_TSI_DisableIT_EndOfSequence(TSI);

    /* Back to continuous scan */
    FL_TSI_SetScanMode(TSI, FL_TSI_SCAN_MODE_CONTINUOUS);
    
    /* Init widgets */
    TSI_Widget_InitAll();
    
    /* Restore TSI state */
    if(oldLibStat == TSI_LIB_RUNNING)
    {
        TSI_Resume();
    }
    
    return TSI_PASS;
}

/* Default error handler */
__WEAK void TSI_ErrorHandler(void* obj, uint32_t val, const char* msg)
{
    (void)obj;
    (void)val;
    (void)msg;

    while (1)
    {
        /* Do nothing */
    };
}

/********************** (C) COPYRIGHT Fudan Microelectronics **** END OF FILE ***********************/
