/**
  ****************************************************************************************************
  * @file    tsi_baseline.c
  * @author  FMSH Application Team
  * @brief   Source file of FL Driver Library
  ****************************************************************************************************
  * @attention
  *
  * Copyright (c) [2021] [Fudan Microelectronics]
  * THIS SOFTWARE is licensed under Mulan PSL v2.
  * You can use this software according to the terms and conditions of the Mulan PSL v2.
  * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
  * EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
  * MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
  * See the Mulan PSL v2 for more details.
  *
  ****************************************************************************************************
  */
/* Includes ----------------------------------------------------------------------------------------*/
#include "tsi_baseline.h"

#if (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_IIR_BASELINE)
#include "tsi_filter.h"
#endif

/* Macros ---------------------------------------------------------------------------------------------*/
/* Bucket status defines */
#define TSI_BUCKET_STATUS_INIT              (0U)
#define TSI_BUCKET_STATUS_NOISE             (1U)
#define TSI_BUCKET_STATUS_ERROR             (2U)
#define TSI_BUCKET_STATUS_NEG_NOISE         (3U)
#define TSI_BUCKET_STATUS_NEG_ERROR         (4U)

/* Defines --------------------------------------------------------------------------------------------*/

/* Exported Functions ---------------------------------------------------------------------------------*/
void TSI_Baseline_Init(TSI_SensorTypeDef* sensor)
{
    sensor->baseline = sensor->rawCount;

#if (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_NORMAL_BASELINE) 

    sensor->lowBaselineCount = 0U;
    sensor->bucketCount = 0U;
    sensor->bucketStatus = TSI_BUCKET_STATUS_INIT;
    
#elif (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_IIR_BASELINE) 

    TSI_Filter_IIRInit(&sensor->iirBuffer, sensor->rawCount);
    
#endif
}

void TSI_Baseline_Update(TSI_SensorTypeDef* sensor)
{
    TSI_WidgetTypeDef* widget = sensor->meta->parent;
    int32_t diffCount = (int32_t)sensor->rawCount - (int32_t)sensor->baseline;
    
#if (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_NORMAL_BASELINE) 
    if(sensor->rawCount >= sensor->baseline)
    {
        sensor->lowBaselineCount = 0;
    }
    
    /* Reset baseline if sensor signal has been lower than negNoiseTh for lowBaselineRstCount counts */
    if(diffCount < -widget->negNoiseTh)
    {
        if(sensor->lowBaselineCount >= widget->lowBaselineRstCount)
        {
            TSI_Baseline_Init(sensor);
        }
        else
        {
            sensor->lowBaselineCount++;
        }
    }
    else
    {
    #if (!TSI_SENSOR_AUTO_RESET_EN)
        /* 
           When sensor's auto-reset is disabled, baseline only 
           updates when diffCount < noiseTh 
        */
        if(diffCount <= widget->noiseTh)
        {
    #endif

        if(diffCount <= widget->noiseTh && diffCount >= 0)
        {
            /* diffCount belongs to [0, noiseTh] */
            if(sensor->bucketStatus != TSI_BUCKET_STATUS_NOISE)
            {
                sensor->bucketStatus = TSI_BUCKET_STATUS_NOISE;
                sensor->bucketCount = 0;
            }

            sensor->bucketCount += diffCount;
            if(sensor->bucketCount > widget->noiseBucket)
            {
                sensor->baseline++;
                sensor->bucketCount = 0;
            }
        }
        else if(diffCount >= -widget->negNoiseTh && diffCount < 0)
        {
            /* diffCount belongs to [-negNoiseTh, 0) */
            if(sensor->bucketStatus != TSI_BUCKET_STATUS_NEG_NOISE)
            {
                sensor->bucketStatus = TSI_BUCKET_STATUS_NEG_NOISE;
                sensor->bucketCount = 0;
            }

            sensor->bucketCount += diffCount;
            if(sensor->bucketCount > widget->negNoiseBucket)
            {
                sensor->baseline--;
                sensor->bucketCount = 0;
            }
        }
        else if(diffCount > widget->noiseTh && diffCount <= widget->activeTh)
        {
            /* diffCount belongs to (noiseTh, activeTh] */
            if(sensor->bucketStatus != TSI_BUCKET_STATUS_ERROR)
            {
                sensor->bucketStatus = TSI_BUCKET_STATUS_ERROR;
                sensor->bucketCount = 0;
            }

            sensor->bucketCount++;
            if(sensor->bucketCount > widget->errorStepCount)
            {
                sensor->baseline += widget->errorStepSize;
                sensor->bucketCount = 0;
            }
        }
        else
        {
            /* Do nothing */
        }
        
    #if (!TSI_SENSOR_AUTO_RESET_EN)
        }
    #endif
    }
    
#elif (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_IIR_BASELINE)

    /* IIR baseline */
#if (!TSI_SENSOR_AUTO_RESET_EN)
    /* 
       When sensor's auto-reset is disabled, baseline only 
       updates when diffCount <= noiseTh 
    */
    if(diffCount <= widget->noiseTh)
    {
#endif
        
        uint16_t m_rawCount = sensor->rawCount;
        TSI_Filter_IIRUpdate(&sensor->iirBuffer, widget->baselineIIRCoeff, &m_rawCount);
        sensor->baseline = m_rawCount;
        
#if (!TSI_SENSOR_AUTO_RESET_EN)
    }
#endif
   
#endif
    
    /* Calculate diffCount using new baseline value */
    diffCount = (int32_t)sensor->rawCount - (int32_t)sensor->baseline;
    
    /* If proximity sensor and filter is enabled, we should update its mode */
    if (widget->meta->type == TSI_WIDGET_PROXIMITY)
    {
        TSI_ProximityWidgetTypeDef* proxWidget = (TSI_ProximityWidgetTypeDef*)widget;
        
        if (proxWidget->filterMode == 0 && diffCount > proxWidget->filterActiveTh)
        {
            proxWidget->filterMode = 1;
        }
        else if(proxWidget->filterMode == 1 && diffCount < proxWidget->filterDetectTh)
        {
            proxWidget->filterMode = 0;
        }
    }

    /* Update sensor diffCount */
    if(diffCount > widget->noiseTh)
    {
        sensor->diffCount = sensor->rawCount - sensor->baseline;
    }
    else
    {
        sensor->diffCount = 0;
    }
}

/********************** (C) COPYRIGHT Fudan Microelectronics **** END OF FILE ***********************/
