/**
  ****************************************************************************************************
  * @file    tsi_calibration.c
  * @author  FMSH Application Team
  * @brief   Source file of FL Driver Library
  ****************************************************************************************************
  * @attention
  *
  * Copyright (c) [2021] [Fudan Microelectronics]
  * THIS SOFTWARE is licensed under Mulan PSL v2.
  * You can use this software according to the terms and conditions of the Mulan PSL v2.
  * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
  * EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
  * MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
  * See the Mulan PSL v2 for more details.
  *
  ****************************************************************************************************
  */
/* Includes ----------------------------------------------------------------------------------------*/
#include "tsi_calibration.h"
#include "tsi_widget.h"
#include "tsi_sensor.h"
#include "tsi.h"

/* Macros ---------------------------------------------------------------------------------------------*/

/* Defines --------------------------------------------------------------------------------------------*/

/* Exported Functions ---------------------------------------------------------------------------------*/
TSI_RetCodeType TSI_CalibrateWidgetIDAC(TSI_WidgetTypeDef *widget, uint8_t targetPercent)
{
    /* 
        We use successive approach method to calibrate widget idacMod.
        If compensation IDAC is enabled, we keep sensor idacComp equals
        to idacMod, which producing the best performance.
        
        Calibration steps:
        1. Enumerate through widget sensors, calibrating widget idacMod to make 
           sensor rawCount approaching { targetPercent } percent of maximum 
           rawCount. 
        2. Find the maximum idacMod among those calibration results. It is the
           result we want. 
    */
    uint8_t maxIdacMod = 0U;
    uint16_t target = targetPercent * ((0x1U << widget->resolution) - 1U) / 100U;
    const TSI_MetaWidgetTypeDef *metaWidget = widget->meta;
    TSI_SensorTypeDef *sensor = &metaWidget->sensors[0];
    int i;
    uint32_t timeout;

    /* Save rumtime context -------------------------------------------------*/
    TSI_LibStatType libStat = (TSI_LibStatType)TSI_RDSTAT_LIB_STAT();
    uint8_t widgetEnableStat = widget->enable;
    
    /* Auto set scan period */
    TSI_AutoSetScanPeriod();

    /* Set single scan mode -------------------------------------------------*/
    if (libStat == TSI_LIB_RUNNING)
    {
        /* Suspend TSI library before calibration */
        (void)TSI_Suspend();
    }
    else if (libStat == TSI_LIB_RESET)
    {
        /* Cannot perform a calibration upon reset */
        return TSI_UNSUPPORTED;
    }

    /* Set to single scan mode */
    FL_TSI_SetScanMode(TSI, FL_TSI_SCAN_MODE_SINGLE);
    
    /* Force single IDAC mode */
    for(i=0; i<metaWidget->sensorCnt; i++)
    {
        metaWidget->sensors[i].idacComp = 0U;
    }
    
    /* Calibrate sensors ----------------------------------------------------*/
    for (i = 0; i < metaWidget->sensorCnt; i++)
    {
        uint8_t bitMask = 0x1U << (TSI_IDAC_BITWIDTH - 1U);
        widget->idacMod = 0U;
        
        /* Ensure the sensor is enabled */
        TSI_Sensor_Enable(sensor);

        /* Successive approach method */
        while (bitMask != 0U)
        {
            /* Update idacMod */
            widget->idacMod |= bitMask;

            /* Apply configuration */
            TSI_Sensor_ApplyConfig(sensor);

            /* Perform a single scan */
            (void)TSI_Start();
            timeout = TSI_CALIB_TIMEOUT_VALUE;
            while (!TSI_GETSTAT_SCAN_CPLT() && timeout > 0)
            {
                timeout--;
            }
            if (timeout == 0)
            {
                return TSI_TIMEOUT;
            }
            (void)TSI_Suspend();

            /* Adjust widget idacMod */
            if (sensor->rawCount < target)
            {
                widget->idacMod &= (~bitMask);
            }

            /* Move to next bit */
            bitMask >>= 1U;
        }

        /* idacMod cannot be 0, set to 1 */
        if (widget->idacMod == 0U)
        {
            widget->idacMod = 1U;
        }

#if (TSI_USE_COMP_IDAC)
        /* Store calibration result to idacComp, we use this value to calculate
           final idacComp */
        sensor->idacComp = widget->idacMod;
#endif
        
        /* Update maximum idacMod */
        if (maxIdacMod < widget->idacMod)
        {
            maxIdacMod = widget->idacMod;
        }

        sensor++;
    }

#if (TSI_USE_COMP_IDAC)
    /* set widget idacMod = (maxIdacMod + 1) / 2 */
    widget->idacMod = (maxIdacMod + 1) >> 1;
    
    sensor = &metaWidget->sensors[0];
    for (i = 0; i < metaWidget->sensorCnt; i++, sensor++)
    {
        /* set sensor idacComp = ceil(([calibration idac result] - idacMod) * 85%) */
        sensor->idacComp = (uint8_t)(((sensor->idacComp - widget->idacMod) * 
            TSI_CALIB_TARGET_PERCENTAGE + 99U) / 100U);
        
        /* Apply configuration */
        TSI_Sensor_ApplyConfig(sensor);
    }
#else
    /* Set widget idacMod = maxIdacMod */
    widget->idacMod = maxIdacMod;
#endif

    /* Perform a single scan to refresh rawCount */
    (void)TSI_Start();
    timeout = TSI_CALIB_TIMEOUT_VALUE;
    while (!TSI_GETSTAT_SCAN_CPLT() && timeout > 0)
    {
        timeout--;
    }
    if (timeout == 0)
    {
        return TSI_TIMEOUT;
    }
    TSI_CLRSTAT_SCAN_CPLT();
    (void)TSI_Suspend();

    /* Restore rumtime context ----------------------------------------------*/
    if (widgetEnableStat == 0U)
    {
        TSI_Widget_Disable(widget);
    }

    FL_TSI_SetScanMode(TSI, FL_TSI_SCAN_MODE_CONTINUOUS);

    if (libStat == TSI_LIB_INIT || libStat == TSI_LIB_RUNNING)
    {
        (void)TSI_Start();
    }
    
    return TSI_PASS;
}

TSI_RetCodeType TSI_CalibrateAllWidgets(void)
{
    int i;
    for (i = 0; i < TSI_TOTAL_WIDGETS_NUM; i++)
    {
        TSI_CalibrateWidget(TSI_WidgetPointers[i]);
    }

    return TSI_PASS;
}

TSI_RetCodeType TSI_CalibrateWidget(TSI_WidgetTypeDef *widget)
{
    uint16_t realIdacMod, tmpResolution;
    uint32_t minSmplClkDiv, tmpClkDiv, tmpN, maxSmplClkFreq;
    uint8_t maxCs = 0U;
    uint64_t tmpMaxCs;
    
    /*
        Before calibration, we should calculate the initial TSI sample 
        clock divider TSI_SW_DIV based on TSI frequency.
    */
    
    /* Calculate minimum TSI_SW_DIV */
    minSmplClkDiv = (TSI_Module.hardwareCfg.opClkFreq / TSI_MAX_SMPL_CLK) + 1;
    
    /* Find and set the initial TSI_SW_DIV to widget */
    for (tmpClkDiv = 0U; ((uint32_t)1UL << tmpClkDiv) < minSmplClkDiv; tmpClkDiv++)
    {
        /* Do Nothing */
    }
    tmpClkDiv = 1U << tmpClkDiv;
    if (tmpClkDiv > TSI_MAX_SMPL_CLK_DIV)
    {
        /* Call error handler */
        TSI_ErrorHandler(widget, 0U, "No valid sample clock divisor!");
        return TSI_OPERATION_ERR;
    }
    widget->sampleClkDiv = tmpClkDiv;

    /* 
        1. Perfrom a widget IDAC calibration, setting target to { TSI_CALIB_TARGET_PERCENTAGE }% of rawCount 
           maximum using following configuration:
           - set resolution to { TSI_CALIB_INITIAL_RESOLUTION }
           - select PRS as sample clock source
        2. Calculate the sensor's static capcitance Cs using the calibration result
        3. Get the sensor's sample clock divider based on Cs  
        4. Get the sensor's resolution based on other params and user-input sensitivity
    */
    
    /* Set configurations and perform calibration */
    widget->resolution = TSI_CALIB_INITIAL_RESOLUTION;
    FL_TSI_SetSampleClockSource(TSI, FL_TSI_SAMPLE_CLK_SOURCE_PRS_CLK);
    TSI_CalibrateWidgetIDAC(widget, TSI_CALIB_TARGET_PERCENTAGE);
    
    /* Calculate sensor's Cs(round up) */
#if (TSI_USE_COMP_IDAC)
    realIdacMod = widget->idacMod << 1U;
#else
    realIdacMod = widget->idacMod;
#endif

    /* Calculate sensor's Cs(round up) */
    tmpMaxCs = ((uint64_t)realIdacMod * 1200000U * 85U) / (TSI_Module.hardwareCfg.opClkFreq / (widget->sampleClkDiv << 2U));
    maxCs = (uint8_t)(tmpMaxCs / 100U + ((tmpMaxCs % 100U) > 0U ? 1U : 0U));

    if (maxCs > TSI_MAX_CS || maxCs < TSI_MIN_CS)
    {
        /* Call error handler */
        TSI_ErrorHandler(widget, maxCs, "Sensor Cs out of range!");
        return TSI_OPERATION_ERR;
    }

    /* Calculate maximum sample clock frequency */
    maxSmplClkFreq = (uint32_t)((uint64_t)200000000000U / (TSI_SERIES_RESISTOR_TOTAL * maxCs));
    if(maxSmplClkFreq > TSI_MAX_SMPL_CLK)
    {
        maxSmplClkFreq = TSI_MAX_SMPL_CLK;
    }

    /* Calculate minimum TSI_SW_DIV */
    minSmplClkDiv = (TSI_Module.hardwareCfg.opClkFreq / maxSmplClkFreq) + 1;

    // Find and set the TSI_SW_DIV to widget
    for (tmpClkDiv = 0U; ((uint32_t)1UL << tmpClkDiv) < minSmplClkDiv; tmpClkDiv++)
    {
        /* Do Nothing */
    }
    tmpClkDiv = 1U << tmpClkDiv;
    if (tmpClkDiv > TSI_MAX_SMPL_CLK_DIV)
    {
        /* Call error handler */
        TSI_ErrorHandler(widget, 0U, "No valid sample clock divisor!");
        return TSI_OPERATION_ERR;
    }
    widget->sampleClkDiv = tmpClkDiv;

    // Calculate resolution
    tmpN = ((uint64_t)maxCs * 1000U * widget->sensitivity) / TSI_CALIB_TARGET_PERCENTAGE + 1U;
    tmpResolution = TSI_MIN_RESOLUTION;
    while (((uint32_t)1UL << tmpResolution) < tmpN)
    {
        tmpResolution++;
    }

    if (tmpResolution > TSI_MAX_RESOLUTION)
    {
        /* Call error handler */
        TSI_ErrorHandler(widget, tmpResolution, "Resolution out of range!");
        return TSI_OPERATION_ERR;
    }

    // set the resolution to widget
    widget->resolution = tmpResolution;

    /*
        After calibration, we should perform an IDAC calibration, which also
        refreshes the rawCount.
    */
    TSI_CalibrateWidgetIDAC(widget, TSI_CALIB_TARGET_PERCENTAGE);

    return TSI_PASS;
}

TSI_RetCodeType TSI_AutoSetScanPeriod(void)
{
    uint64_t scanTime = 0U;
    uint64_t opClkFreq = TSI_Module.hardwareCfg.opClkFreq /
                         TSI_Module.hardwareCfg.opClkPrescaler;
    int i;
    
    /* Calculate total scan time */
    for (i = 0; i < TSI_TOTAL_WIDGETS_NUM; i++)
    {
        const TSI_WidgetTypeDef* widget = TSI_WidgetPointers[i];
        
        /* 
            Scan time for one widget = 1/f_opclk * (2^N) * n_sns 
                                       + t_smpl + t_intv
         */
        scanTime += ((uint64_t)1000000U << widget->resolution) 
                    * widget->meta->sensorCnt / opClkFreq;
        scanTime += (TSI_Module.scanCfg.sampleIdleTime + 1) << 5U;
        scanTime += (TSI_Module.scanCfg.sampleInterval + 1) << 5U;
    }
    
    /* Select an appropriate scan period */
    FL_TSI_ScanPeriodInit(scanTime / 1000U);
    
    return TSI_PASS;
}

/********************** (C) COPYRIGHT Fudan Microelectronics **** END OF FILE ***********************/
