/**
  ****************************************************************************************************
  * @file    tsi_filter.c
  * @author  FMSH Application Team
  * @brief   Source file of FL Driver Library
  ****************************************************************************************************
  * @attention
  *
  * Copyright (c) [2021] [Fudan Microelectronics]
  * THIS SOFTWARE is licensed under Mulan PSL v2.
  * You can use this software according to the terms and conditions of the Mulan PSL v2.
  * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
  * EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
  * MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
  * See the Mulan PSL v2 for more details.
  *
  ****************************************************************************************************
  */
/* Includes ----------------------------------------------------------------------------------------*/
#include "tsi_filter.h"

/* Macros ---------------------------------------------------------------------------------------------*/

/* Defines --------------------------------------------------------------------------------------------*/
/* 高级二段式IIR滤波器系数矩阵 */
TSI_PROX_FILTER_ADVIIR_COEFF_PERF_QP
TSI_PROX_FILTER_ADVIIR_COEFF_PERF
TSI_PROX_FILTER_ADVIIR_COEFF_FAST_QP
TSI_PROX_FILTER_ADVIIR_COEFF_FAST

/* Exported Functions ---------------------------------------------------------------------------------*/
void TSI_Filter_InitSensor(TSI_SensorTypeDef* sensor)
{
#if (TSI_REGULAR_FILTER_EN || TSI_PROX_FILTER_EN)
    const TSI_MetaSensorTypeDef* metaSensor = sensor->meta;
    uint16_t initVal = sensor->rawCount;
    
    if(metaSensor->filterType == TSI_FILTER_REGULAR_SNS)
    {
    #if TSI_REGULAR_FILTER_EN
        TSI_RegularSnsFilterTypeDef* regFilter = (TSI_RegularSnsFilterTypeDef*)(metaSensor->filter);
    #if TSI_REGULAR_FILTER_MEDIAN_EN
        TSI_Filter_Median3Init(regFilter->medianBuffer, initVal);
    #endif  /* TSI_REGULAR_FILTER_MEDIAN_EN */
    #if TSI_REGULAR_FILTER_IIR_EN
        TSI_Filter_IIRInit(&regFilter->iirBuffer, initVal);
    #endif  /* TSI_REGULAR_FILTER_IIR_EN */
    #if TSI_REGULAR_FILTER_AVERAGE_EN
        TSI_Filter_Average4Init(regFilter->averageBuffer, initVal);
    #endif  /* TSI_REGULAR_FILTER_AVERAGE_EN */
    #endif  /* TSI_REGULAR_FILTER_EN */
    }
    else if(metaSensor->filterType == TSI_FILTER_PROX_SNS)
    {
    #if TSI_PROX_FILTER_EN
        TSI_ProxSnsFilterTypeDef* proxFilter = (TSI_ProxSnsFilterTypeDef*)(metaSensor->filter);
    #if TSI_PROX_FILTER_MEDIAN_EN
        TSI_Filter_Median3Init(proxFilter->medianBuffer, initVal);
    #endif  /* TSI_PROX_FILTER_MEDIAN_EN */
    #if TSI_PROX_FILTER_ADVIIR_EN
        TSI_Filter_ADVIIRInit(proxFilter->adviirBuffer, initVal);
    #endif  /* TSI_PROX_FILTER_IIR_EN */
    #if TSI_PROX_FILTER_AVERAGE_EN
        TSI_Filter_Average4Init(proxFilter->averageBuffer, initVal);
    #endif  /* TSI_PROX_FILTER_AVERAGE_EN */
    #endif  /* TSI_PROX_FILTER_EN */
    }
    else
    {
        /* Do nothing */
    }
# endif
}

void TSI_Filter_UpdateSensor(TSI_SensorTypeDef* sensor)
{
#if (TSI_REGULAR_FILTER_EN || TSI_PROX_FILTER_EN)
    const TSI_MetaSensorTypeDef* metaSensor = sensor->meta;
    uint16_t *input = &sensor->rawCount;
    
    if(metaSensor->filter != NULL)
    {
        if(metaSensor->filterType == TSI_FILTER_REGULAR_SNS)
        {
        #if TSI_REGULAR_FILTER_EN
            TSI_RegularSnsFilterTypeDef* regFilter = (TSI_RegularSnsFilterTypeDef*)(metaSensor->filter);
        #if TSI_REGULAR_FILTER_MEDIAN_EN
            TSI_Filter_Median3Update(regFilter->medianBuffer, input);
        #endif  /* TSI_REGULAR_FILTER_MEDIAN_EN */
        #if TSI_REGULAR_FILTER_IIR_EN
            TSI_Filter_IIRUpdate(&regFilter->iirBuffer, TSI_REGULAR_FILTER_IIR_COEF, input);
        #endif  /* TSI_REGULAR_FILTER_IIR_EN */
        #if TSI_REGULAR_FILTER_AVERAGE_EN
            TSI_Filter_Average4Update(regFilter->averageBuffer, input);
        #endif  /* TSI_REGULAR_FILTER_AVERAGE_EN */
        #endif  /* TSI_REGULAR_FILTER_EN */
        }
        else if(metaSensor->filterType == TSI_FILTER_PROX_SNS)
        {
        #if TSI_PROX_FILTER_EN
            TSI_ProxSnsFilterTypeDef* proxFilter = (TSI_ProxSnsFilterTypeDef*)(metaSensor->filter);
        #if TSI_PROX_FILTER_MEDIAN_EN
            TSI_Filter_Median3Update(proxFilter->medianBuffer, input);
        #endif  /* TSI_PROX_FILTER_MEDIAN_EN */
        #if TSI_PROX_FILTER_ADVIIR_EN
            TSI_Filter_ADVIIRUpdate(proxFilter->adviirBuffer, input, 
                                    ((TSI_ProximityWidgetTypeDef*)metaSensor->parent)->filterMode);
        #endif  /* TSI_PROX_FILTER_IIR_EN */
        #if TSI_PROX_FILTER_AVERAGE_EN
            TSI_Filter_Average4Update(proxFilter->averageBuffer, input);
        #endif  /* TSI_PROX_FILTER_AVERAGE_EN */
        #endif  /* TSI_PROX_FILTER_EN */
        }
        else
        {
            /* Do nothing */
        }
    }
    else
    {
        /* Do nothing */
    }
# endif
}

/* 4阶平均滤波器 --------------------------------------------------------------*/
void TSI_Filter_Average4Init(uint16_t *buffer, uint16_t initVal)
{
    buffer[0] = buffer[1] = buffer[2] = initVal;
}

void TSI_Filter_Average4Update(uint16_t *buffer, uint16_t* input)
{
    uint32_t result = (buffer[0] + buffer[1] + buffer[2] + *input) >> 2;
    buffer[0] = buffer[1];
    buffer[1] = buffer[2];
    buffer[2] = *input;
    *input = (uint16_t)(result & 0xFFFFU);
}

/* 3阶中值滤波器 --------------------------------------------------------------*/
void TSI_Filter_Median3Init(uint16_t *buffer, uint16_t initVal)
{
    buffer[0] = buffer[1] = buffer[2] = initVal;
}

void TSI_Filter_Median3Update(uint16_t *buffer, uint16_t* input)
{
    uint16_t tmpA, tmpB;

    tmpA = buffer[0];
    tmpB = buffer[1];
    if(buffer[0] < buffer[1])
    {
        tmpA = buffer[1];
        tmpB = buffer[0];
    }
    if(tmpA > *input)
    {
        tmpA = *input > tmpB ? *input : tmpB;
    }

    buffer[0] = buffer[1];
    buffer[1] = *input;
    *input = (uint16_t)(tmpA & 0xFFFFU);
}

/* 抖动滤波器 ----------------------------------------------------------------*/
void TSI_Filter_JitterInit(uint16_t *buffer, uint16_t initVal)
{
    buffer[0] = initVal;
}

void TSI_Filter_JitterUpdate(uint16_t *buffer, uint16_t* input)
{
    uint16_t tmp = *input;

    if(buffer[0] + 1 < tmp)
    {
        tmp--;
    }
    else if(buffer[0] - 1 > tmp)
    {
        tmp++;
    }
    else
    {
        /* Do nothing */
    }
    
    buffer[0] = *input;
    *input = (uint16_t)(tmp & 0xFFFFU);
}

/* 1阶IIR滤波器 ---------------------------------------------------------------*/
void TSI_Filter_IIRInit(uint32_t *buffer, uint16_t initVal)
{
    buffer[0] = (uint32_t)initVal << 7U;
}

void TSI_Filter_IIRUpdate(uint32_t *buffer, uint8_t coef, uint16_t* input)
{
    uint32_t tmpQ7;
    uint32_t inputQ7 = ((uint32_t)*input) << 7U; 
    
    /*
        buffer[0]: Y(N-1) ==> Fix-Point Number(Q7)
        y(N) = (coef * x(N) + (256 - coef) * y(N-1)) / K, K = 256
    */
    tmpQ7 = ((uint64_t)coef * inputQ7 + (uint64_t)(256 - coef) * buffer[0]) >> 8U;
    buffer[0] = tmpQ7;
    *input = tmpQ7 >> 7U;
}

/* 高级二段式IIR滤波器（接近感应使用） -----------------------------------------*/
void TSI_Filter_ADVIIRInit(uint32_t *buffer, uint16_t initVal)
{
    int i;
    int32_t *filterBuf = (int32_t*)buffer;
    
    for (i=0; i<6; i++)
    {
        filterBuf[i] = ((int32_t)initVal) << 14U;
    }
}

void TSI_Filter_ADVIIRUpdate(uint32_t *buffer, uint16_t* input, uint8_t filterMode)
{
    int32_t tmpQ14;
    uint32_t restmpQ14;
    int32_t inputQ14 = ((int32_t)*input) << 14U; 
    int32_t *filterBuf = (int32_t*)buffer;
    
    
    /*
        buffer[6]: X(N-1), X(N-2), X(N-3), Y(N-1), Y(N-2), Y(N-3)
        coef[0]: b[4]
        coef[1]: a[4]
    */
    if(filterMode == 0)
    {
        tmpQ14 = - (((int64_t)ADVIIR_COEFF_PERF[1][1] * filterBuf[3]) >> ADVIIR_COEFF_PERF_QP)
                 - (((int64_t)ADVIIR_COEFF_PERF[1][2] * filterBuf[4]) >> ADVIIR_COEFF_PERF_QP)
                 - (((int64_t)ADVIIR_COEFF_PERF[1][3] * filterBuf[5]) >> ADVIIR_COEFF_PERF_QP) 
                 + (((int64_t)ADVIIR_COEFF_PERF[0][0] * inputQ14    ) >> ADVIIR_COEFF_PERF_QP)
                 + (((int64_t)ADVIIR_COEFF_PERF[0][1] * filterBuf[0]) >> ADVIIR_COEFF_PERF_QP) 
                 + (((int64_t)ADVIIR_COEFF_PERF[0][2] * filterBuf[1]) >> ADVIIR_COEFF_PERF_QP) 
                 + (((int64_t)ADVIIR_COEFF_PERF[0][3] * filterBuf[2]) >> ADVIIR_COEFF_PERF_QP);
    }
    else
    {
        tmpQ14 = - (((int64_t)ADVIIR_COEFF_FAST[1][1] * filterBuf[3]) >> ADVIIR_COEFF_FAST_QP)
                 - (((int64_t)ADVIIR_COEFF_FAST[1][2] * filterBuf[4]) >> ADVIIR_COEFF_FAST_QP) 
                 - (((int64_t)ADVIIR_COEFF_FAST[1][3] * filterBuf[5]) >> ADVIIR_COEFF_FAST_QP) 
                 + (((int64_t)ADVIIR_COEFF_FAST[0][0] * inputQ14    ) >> ADVIIR_COEFF_FAST_QP)
                 + (((int64_t)ADVIIR_COEFF_FAST[0][1] * filterBuf[0]) >> ADVIIR_COEFF_FAST_QP) 
                 + (((int64_t)ADVIIR_COEFF_FAST[0][2] * filterBuf[1]) >> ADVIIR_COEFF_FAST_QP)
                 + (((int64_t)ADVIIR_COEFF_FAST[0][3] * filterBuf[2]) >> ADVIIR_COEFF_FAST_QP);
    }
    
    // Result should >= 0
    if(tmpQ14 < 0)
    {
        tmpQ14 = 0;
    }
    
    filterBuf[2] = filterBuf[1];
    filterBuf[1] = filterBuf[0];
    filterBuf[0] = inputQ14;
    filterBuf[5] = filterBuf[4];
    filterBuf[4] = filterBuf[3];
    filterBuf[3] = tmpQ14;
    
    // Result should <= MAX(UINT16)
    restmpQ14 = ((uint32_t)tmpQ14) >> 14U;
    *input = (restmpQ14 > 0xFFFFU) ? 0xFFFFU : restmpQ14;
}

/********************** (C) COPYRIGHT Fudan Microelectronics **** END OF FILE ***********************/
