/**
  *******************************************************************************************************
  * @file    tsi_objects.h
  * @author  FMSH Application Team
  * @brief   Header file of FL Driver Library
  *******************************************************************************************************
  * @attention
  *
  * Copyright (c) [2021] [Fudan Microelectronics]
  * THIS SOFTWARE is licensed under Mulan PSL v2.
  * You can use this software according to the terms and conditions of the Mulan PSL v2.
  * You may obtain a copy of Mulan PSL v2 at:
  *          http://license.coscl.org.cn/MulanPSL2
  * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
  * EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
  * MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
  * See the Mulan PSL v2 for more details.
  *
  *******************************************************************************************************
  */
/* Prevent recursive inclusion ------------------------------------------------------------------------*/
#ifndef __TSI_LIB_OBJECTS_H__
#define __TSI_LIB_OBJECTS_H__

#ifdef __cplusplus
extern "C" {
#endif

/* Includes -------------------------------------------------------------------------------------------*/
#include "tsi_conf.h"

/* Macros ---------------------------------------------------------------------------------------------*/
#define TSI_STAT_SCAN_CPLT              (7U)
#define TSI_STAT_SCAN_CPLT_MSK          (0x1U << TSI_STAT_SCAN_CPLT)

#define TSI_STAT_LIB_STAT               (0U)
#define TSI_STAT_LIB_STAT_MSK           (0xFU << TSI_STAT_LIB_STAT)

#define TSI_SETSTAT_SCAN_CPLT()         (TSI_Module.status |= TSI_STAT_SCAN_CPLT_MSK)
#define TSI_CLRSTAT_SCAN_CPLT()         (TSI_Module.status &= ~TSI_STAT_SCAN_CPLT_MSK)
#define TSI_GETSTAT_SCAN_CPLT()         ((TSI_Module.status & TSI_STAT_SCAN_CPLT_MSK) == TSI_STAT_SCAN_CPLT_MSK) 

#define TSI_WRSTAT_LIB_STAT(__VAL__)    do {    \
                                            TSI_Module.status &= ~TSI_STAT_LIB_STAT_MSK;    \
                                            TSI_Module.status |= ((__VAL__ << TSI_STAT_LIB_STAT) & TSI_STAT_LIB_STAT_MSK);  \
                                        } while(0)   
#define TSI_RDSTAT_LIB_STAT()           ((TSI_Module.status & TSI_STAT_LIB_STAT_MSK) >> TSI_STAT_LIB_STAT)

/* Defines --------------------------------------------------------------------------------------------*/
typedef enum _TSI_WidgetType
{
    TSI_WIDGET_BUTTON       = 0,
    TSI_WIDGET_PROXIMITY    = 1,
    TSI_WIDGET_SLIDER       = 2,
} TSI_WidgetType;

typedef enum _TSI_LibStatType
{
    TSI_LIB_RESET       = 0,
    TSI_LIB_INIT        = 1,
    TSI_LIB_RUNNING     = 2,
    TSI_LIB_SUSPEND     = 3,
} TSI_LibStatType;

typedef struct _TSI_Hardware
{
    /*! Operation clock source */
    uint32_t opClkSource;
    
    /*! Operation clock prescaler */
    uint32_t opClkPrescaler;

    /*! Operation clock real frequency */
    uint32_t opClkFreq;

} TSI_HardwareTypeDef;

typedef struct _TSI_Scan
{
    /*! Scan period */
    uint8_t scanPeriod;

    /*! Scan interval */
    uint8_t sampleInterval;

    /*! Scan prepare(idle) time */
    uint8_t sampleIdleTime;

} TSI_ScanTypeDef;

typedef union _TSI_Pin
{
    struct 
    {
        /*! [3:0] pin id */
        uint16_t pin:4;

        /*! [9:4] port id */
        uint16_t port:6;

        /* ! [15:10] channel id */
        uint16_t channel:6;
    } pinConfig;

    uint16_t pinUniqueId;
    
} TSI_PinTypeDef;

typedef struct _TSI_Module
{
    /*! Firmware ID for revision checking */ 
    uint32_t firmwareId;

    /*! Debug command storage */
    uint32_t debugCmd;

    /*! Hardware related configurations */
    TSI_HardwareTypeDef hardwareCfg;

    /*! Scan related configurations */
    TSI_ScanTypeDef scanCfg;

    /*! Lib status */
    uint8_t status;

} TSI_ModuleTypeDef;

typedef enum _TSI_FilterType
{
    TSI_FILTER_REGULAR_SNS  = 0,
    TSI_FILTER_PROX_SNS     = 1,
} TSI_FilterType;

typedef struct _TSI_MetaSensor
{
    /*! Sensor parent widget */
    struct _TSI_Widget* parent;

    /*! Sensor channel */
    uint8_t channel;

    /*! Filter buffer */
    void* filter;

    /*! Filter type */
    TSI_FilterType filterType;

    /*! Debounce counters array pointer */
    uint8_t *debArray;

} TSI_MetaSensorTypeDef;

typedef struct _TSI_Sensor
{
    /*! Meta info */
    const TSI_MetaSensorTypeDef* const meta;

    /*--- Hardware Parameters ---*/
    /*! Compensation IDAC current */ 
    uint16_t idacComp;

    /*--- Values ---*/
    /*! Raw count */
    uint16_t rawCount;

    /*! Baseline */
    uint16_t baseline;

    /*! actual signal */
    uint16_t diffCount;

    /*--- Algorithm variables ---*/
#if (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_NORMAL_BASELINE) 
    /*! Sensor bucket counter */
    uint16_t bucketCount;

    /*! Sensor reset counter */
    uint16_t lowBaselineCount;

    /*! Sensor bucket status */
    uint8_t bucketStatus;

#elif (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_IIR_BASELINE) 
    /*! Baseline IIR Buffer */
    uint32_t iirBuffer;
    
    /*! Padding, useless */
    uint8_t padding;
    
#endif

    /*! Sensor active status */
    uint8_t status;     /** bit 0: active flag
                            bit 1: proximity flag(Proximity widget sensor(s) only)   */

} TSI_SensorTypeDef;

typedef struct _TSI_MetaWidget
{
    /*! Widget type */
    TSI_WidgetType type;

    /*! Sensors array pointer */
    TSI_SensorTypeDef* const sensors;

    /*! Sensors array count */
    uint8_t sensorCnt;
    
} TSI_MetaWidgetTypeDef;

typedef struct _TSI_Widget
{
    /*! Meta info */
    const TSI_MetaWidgetTypeDef* const meta;
    
    /*! Enable status */
    uint8_t enable;

    /*--- Hardware Parameters ---*/
    /*! Scan Resolution */
    uint8_t resolution;

    /*! Fsw Clock Division */
    uint16_t sampleClkDiv;

    /*! Modulation IDAC current */
    uint16_t idacMod;

    /*--- Software Parameters ---*/
    /*! Active threshold */
    uint16_t activeTh;
    
    /*! Active threshold hysteresis */
    uint16_t activeHys;

    /*! Baseline Noise: threshold */
    uint16_t noiseTh;

#if (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_NORMAL_BASELINE)

    /*! Baseline Negative Noise: threshold */
    uint16_t negNoiseTh;

    /*! Baseline Noise: bucket size */
    uint16_t noiseBucket;

    /*! Baseline Error: Counts before increasing baseline errorStepVal steps */
    uint8_t errorStepCount;

    /*! Baseline Error: step size */
    uint8_t errorStepSize;

    /*! Baseline Negative noise: bucket size */
    uint16_t negNoiseBucket;

    /*! Baseline Negative Error: bucket size */
    uint16_t negErrorBucket;
    
    /*! Counts before reseting baseline when rawCount falls below negNoiseTh */ 
    uint16_t lowBaselineRstCount;

#elif (TSI_SENSOR_BASELINE_TYPE == TSI_SENSOR_IIR_BASELINE) 
    
    /*! Baseline IIR: coefficient */
    uint8_t baselineIIRCoeff;
    
    /*! Padding, useless */
    uint32_t padding[2];

#endif

    /*! On debounce */
    uint8_t onDebounce;

    /*! Off debounce */
    uint8_t offDebounce;

    /*! Calibration sensitivity */
    uint16_t sensitivity;

    /*--- Values ---*/
    /*! Widget active status */
    uint8_t status;         

} TSI_WidgetTypeDef;

#ifdef TSI_WIDGET_BUTTON_USED

typedef struct _TSI_ButtonWidget
{
    /*--- Parameters ---*/
    /*! Base widget */
    TSI_WidgetTypeDef base;
    
    /*--- Values ---*/
    /** Button active status */
    uint8_t buttonStatus;   /** 0: inactive 
                                1: active       */
    
} TSI_ButtonWidgetTypeDef;

#endif

#ifdef TSI_WIDGET_PROXIMITY_USED

typedef struct _TSI_ProximityWidget
{
    /*--- Parameters ---*/
    /*! Base widget */
    TSI_WidgetTypeDef base;
    
    /*! Proximity threshold */
    uint16_t proxTh;
    
    /*! Proximity hysteresis */
    uint16_t proxHys;
    
    /*! Threshold when proximity filter enter active mode */
    uint16_t filterActiveTh;
    
    /*! Threshold when proximity filter enter detect mode, can be negative */
    int16_t filterDetectTh;
    
    /*--- Algorithm variables ---*/
    /*! Proximity filter mode */
    uint8_t filterMode;
    
    /*--- Values ---*/
    /** Proximity active status */
    uint8_t proximityStatus;    /** 0: inactive 
                                    1: proximity  
                                    2: active       */
    
} TSI_ProximityWidgetTypeDef;

#endif

#ifdef TSI_WIDGET_SLIDER_USED

typedef struct _TSI_SliderWidget
{
    /*--- Parameters ---*/
    /*! Base widget */
    TSI_WidgetTypeDef base;
    
    /*--- Values ---*/
    /** Proximity active status */
    uint8_t sliderStatus;       /** 0: inactive  
                                    1: active       */

    /** Slider touch center position 0 - 255 */
    uint8_t centerPos;
    
} TSI_SliderWidgetTypeDef;

#endif

#if TSI_REGULAR_FILTER_EN

typedef struct _TSI_RegularSnsFilter
{
#if TSI_REGULAR_FILTER_AVERAGE_EN
    uint16_t averageBuffer[TSI_REGULAR_FILTER_AVERAGE_ORDER];
#endif /* TSI_REGULAR_FILTER_AVERAGE_EN */

#if TSI_REGULAR_FILTER_MEDIAN_EN
    uint16_t medianBuffer[TSI_REGULAR_FILTER_MEDIAN_ORDER];
#endif /* TSI_REGULAR_FILTER_MEDIAN_EN */

#if TSI_REGULAR_FILTER_IIR_EN
    uint32_t iirBuffer;
#endif /* TSI_REGULAR_FILTER_IIR_EN */

} TSI_RegularSnsFilterTypeDef;

#endif /* TSI_REGULAR_FILTER_EN */

#if TSI_PROX_FILTER_EN

typedef struct _TSI_ProxSnsFilter
{
#if TSI_PROX_FILTER_AVERAGE_EN
    uint16_t averageBuffer[TSI_PROX_FILTER_AVERAGE_ORDER];
#endif /* TSI_PROX_FILTER_AVERAGE_EN */

#if TSI_PROX_FILTER_MEDIAN_EN
    uint16_t medianBuffer[TSI_PROX_FILTER_MEDIAN_ORDER];
#endif /* TSI_PROX_FILTER_MEDIAN_EN */

#if TSI_PROX_FILTER_ADVIIR_EN
    uint32_t adviirBuffer[6];
#endif /* TSI_PROX_FILTER_IIR_EN */

} TSI_ProxSnsFilterTypeDef;

#endif /* TSI_PROX_FILTER_EN */

/* Define widget list and sensor list */
TSI_WIDGET_LIST_DEFINE
TSI_SENSOR_LIST_DEFINE

/**
 * @brief TSI Module 
 */
extern TSI_ModuleTypeDef TSI_Module;

/**
 * @brief TSI Module init value
 */
extern const TSI_ModuleTypeDef TSI_ModuleConstInit;

/**
 * @brief TSI widget list 
 */
extern TSI_WidgetListTypeDef TSI_WidgetList;

/**
 * @brief TSI widget list init value
 */
extern const TSI_WidgetListTypeDef TSI_WidgetListConstInit;

/**
 * @brief TSI sensor list 
 */
extern TSI_SensorListTypeDef TSI_SensorList;

/**
 * @brief TSI sensor list init value
 */
extern const TSI_SensorListTypeDef TSI_SensorListConstInit;

/**
 * @brief TSI channel-to-sensor map
 */
extern TSI_SensorTypeDef* const TSI_ChannelSensorMap[TSI_TOTAL_CHANNEL_NUM];

/**
 * @brief TSI widget pointers
 */
extern TSI_WidgetTypeDef* const TSI_WidgetPointers[TSI_TOTAL_WIDGETS_NUM];

/**
 * @brief TSI widgets meta-info array
 */
extern TSI_MetaWidgetTypeDef const TSI_MetaWidgets[TSI_TOTAL_WIDGETS_NUM];

/**
 * @brief TSI sensors meta-info array
 */
extern TSI_MetaSensorTypeDef const TSI_MetaSensors[TSI_TOTAL_SENSORS_NUM];

/**
 * @brief TSI pins cofiguration
 */
extern TSI_PinTypeDef const TSI_Pins[TSI_TOTAL_PINS_NUM];

/* Exported Functions ---------------------------------------------------------------------------------*/

#ifdef __cplusplus
}
#endif

#endif  /* __TSI_LIB_OBJECTS_H__ */

/********************** (C) COPYRIGHT Fudan Microelectronics **** END OF FILE ***********************/
