/**
  ****************************************************************************************************
  * @file    tsi_sensor.c
  * @author  FMSH Application Team
  * @brief   Source file of FL Driver Library
  ****************************************************************************************************
  * @attention
  *
  * Copyright (c) [2021] [Fudan Microelectronics]
  * THIS SOFTWARE is licensed under Mulan PSL v2.
  * You can use this software according to the terms and conditions of the Mulan PSL v2.
  * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
  * EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
  * MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
  * See the Mulan PSL v2 for more details.
  *
  ****************************************************************************************************
  */
/* Includes ----------------------------------------------------------------------------------------*/
#include "tsi_sensor.h"
#include "tsi_filter.h"
#include "tsi_baseline.h"


/* Macros ---------------------------------------------------------------------------------------------*/
#define TSI_SENSOR_UPDATE_STATUS_ACTIVE     (uint8_t)(0x1U)
#define TSI_SENSOR_UPDATE_STATUS_PROX       (uint8_t)(0x2U)

/* Defines --------------------------------------------------------------------------------------------*/

/* Private Function Prototypes ------------------------------------------------------------------------*/
static void Sensor_UpdateStatus(TSI_SensorTypeDef* sensor, uint8_t type);

/* Exported Functions ---------------------------------------------------------------------------------*/
TSI_RetCodeType TSI_Sensor_ApplyConfig(TSI_SensorTypeDef* sensor)
{
    uint32_t tmp;
    const TSI_MetaSensorTypeDef* metaSensor = sensor->meta;
    TSI_WidgetTypeDef* widget = metaSensor->parent;

    /* Apply sensor channel configurations ----------------------------------*/
    /* Set resolution */
    if(widget->resolution < 8U || widget->resolution > 16U)
    {
        return TSI_PARAM_ERR;
    }
    FL_TSI_SetResolution(TSI, widget->resolution - 8U, metaSensor->channel);
    
    /* Set sample clock prescaler */
    switch(widget->sampleClkDiv)
    {
        case 1: tmp = FL_TSI_SAMPLE_PSC_DIV1; break;
        case 2: tmp = FL_TSI_SAMPLE_PSC_DIV2; break;
        case 4: tmp = FL_TSI_SAMPLE_PSC_DIV4; break;
        case 8: tmp = FL_TSI_SAMPLE_PSC_DIV8; break;
        case 16: tmp = FL_TSI_SAMPLE_PSC_DIV16; break;
        case 32: tmp = FL_TSI_SAMPLE_PSC_DIV32; break;
        
        default:
            return TSI_PARAM_ERR;
    }
    FL_TSI_SetSampleClockPrescaler(TSI, tmp, metaSensor->channel);
    
    /* Set master IDAC current */
    if(widget->idacMod > 127U || widget->idacMod == 0U)
    {
        return TSI_PARAM_ERR;
    }
    FL_TSI_WriteMasterIDACCurrent(TSI, widget->idacMod, metaSensor->channel);

#if (TSI_USE_COMP_IDAC)
    /* Set compensation IDAC current */
    if(sensor->idacComp > 127U)
    {
        return TSI_PARAM_ERR;
    }
    if(sensor->idacComp != 0U)
    {
        FL_TSI_WriteCompensateIDACCurrent(TSI, sensor->idacComp, metaSensor->channel);
        
        /* Set to dual IDAC mode when comensation IDAC current is not 0 */
        FL_TSI_SetIDACMode(TSI, FL_TSI_IDAC_MODE_DUAL, metaSensor->channel);
    }
    else    
    {
        /* Set to single IDAC mode when comensation IDAC current is 0 */
        FL_TSI_SetIDACMode(TSI, FL_TSI_IDAC_MODE_SINGLE, metaSensor->channel);
    }
#else
    /* Set to single IDAC mode */
    FL_TSI_SetIDACMode(TSI, FL_TSI_IDAC_MODE_SINGLE, metaSensor->channel);
#endif
    
    return TSI_PASS;
}

TSI_RetCodeType TSI_Sensor_Init(TSI_SensorTypeDef* sensor)
{   
    /* Init sensor filter buffers */
    TSI_Filter_InitSensor(sensor);

    /* Init sensor baseline */
    TSI_Baseline_Init(sensor);
    
    return TSI_PASS;
}

void TSI_Sensor_Enable(TSI_SensorTypeDef* sensor)
{
    FL_TSI_SetNormSensorChannel(TSI, FL_TSI_GetNormSensorChannel(TSI) | (1 << sensor->meta->channel));
}

void TSI_Sensor_Disable(TSI_SensorTypeDef* sensor)
{
    FL_TSI_SetNormSensorChannel(TSI, FL_TSI_GetNormSensorChannel(TSI) & ~(1 << sensor->meta->channel));
}

TSI_RetCodeType TSI_Sensor_Update(TSI_SensorTypeDef* sensor)
{
    TSI_WidgetTypeDef* widget = sensor->meta->parent;
    
    /* Process rawCount and update baseline -------------*/
    /* Pass rawCount through filter */
    TSI_Filter_UpdateSensor(sensor);

    /* Update baseline with filtered rawCount */
    TSI_Baseline_Update(sensor);
    
    /* Update sensor status -----------------------------*/
    /* Update sensor active status */
    Sensor_UpdateStatus(sensor, TSI_SENSOR_UPDATE_STATUS_ACTIVE);

    if(widget->meta->type == TSI_WIDGET_PROXIMITY)
    {
        /* If it's proximity widget sensor, update sensor proxmity status */
        Sensor_UpdateStatus(sensor, TSI_SENSOR_UPDATE_STATUS_PROX);
    }
    
    return TSI_PASS;
}

/* Private Functions ----------------------------------------------------------------------------------*/
void Sensor_UpdateStatus(TSI_SensorTypeDef* sensor, uint8_t type)
{
    int32_t threshold;
    int32_t diffCount;
    uint16_t hysteresis;
    uint8_t* pDebCount;
    uint8_t newStatus;
    TSI_WidgetTypeDef* widget = sensor->meta->parent;

    /* Get threshold */
    switch (type)
    {
    case TSI_SENSOR_UPDATE_STATUS_ACTIVE:
        threshold = widget->activeTh;
        hysteresis = widget->activeHys;
        pDebCount = &sensor->meta->debArray[0];
        diffCount = sensor->diffCount;
        break;

    case TSI_SENSOR_UPDATE_STATUS_PROX:
        threshold = ((TSI_ProximityWidgetTypeDef*)widget)->proxTh;
        hysteresis = ((TSI_ProximityWidgetTypeDef*)widget)->proxHys;
        pDebCount = &sensor->meta->debArray[1];
        diffCount = (int32_t)sensor->rawCount - (int32_t)sensor->baseline;
        break;
    
    default:
        /* Type not exists, return */
        return;
    }

    /* Add hysteresis to threshold */
    threshold = ((sensor->status & type) == 0U) ? 
            (threshold + hysteresis): 
            (threshold - hysteresis);

    /* Add debounce and update status */
    newStatus = diffCount > threshold ? type : 0U;
    if((newStatus != (sensor->status & type)) && ((*pDebCount) != 0U))
    {
        if(--(*pDebCount) == 0U)
        {
            sensor->status &= ~type;
            sensor->status |= newStatus;
        }
    }
    else
    {
        *pDebCount = ((sensor->status & type) == 0U) ? widget->onDebounce : widget->offDebounce;
    }
}

/********************** (C) COPYRIGHT Fudan Microelectronics **** END OF FILE ***********************/
