/**
  ****************************************************************************************************
  * @file    tsi_widget.c
  * @author  FMSH Application Team
  * @brief   Source file of FL Driver Library
  ****************************************************************************************************
  * @attention
  *
  * Copyright (c) [2021] [Fudan Microelectronics]
  * THIS SOFTWARE is licensed under Mulan PSL v2.
  * You can use this software according to the terms and conditions of the Mulan PSL v2.
  * THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
  * EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
  * MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
  * See the Mulan PSL v2 for more details.
  *
  ****************************************************************************************************
  */
/* Includes ----------------------------------------------------------------------------------------*/
#include "tsi_widget.h"
#include "tsi_sensor.h"

/* Macros ---------------------------------------------------------------------------------------------*/

/* Defines --------------------------------------------------------------------------------------------*/

/* Exported Functions ---------------------------------------------------------------------------------*/
void TSI_Widget_InitAll(void)
{
    int i;
    for (i = 0; i < TSI_TOTAL_WIDGETS_NUM; i++)
    {
        TSI_Widget_Init(TSI_WidgetPointers[i]);
    }
}

void TSI_Widget_Init(TSI_WidgetTypeDef* widget)
{
    int i;
    TSI_SensorTypeDef* sensor = widget->meta->sensors;

    /* 
        TSI_WidgetTypeDef itself only contains parameters and status value, 
        so there's nothing to do with.
    */

    /* Init all sensors */
    for (i = 0; i < widget->meta->sensorCnt; i++)
    {
        TSI_Sensor_Init(sensor);
        sensor++;
    }
}

void TSI_Widget_ApplyConfigToAll(void)
{
    int i;
    for (i = 0; i < TSI_TOTAL_WIDGETS_NUM; i++)
    {
        TSI_Widget_ApplyConfig(TSI_WidgetPointers[i]);
    }
}

void TSI_Widget_EnableAll(void)
{
    int i;
    for (i = 0; i < TSI_TOTAL_WIDGETS_NUM; i++)
    {
        TSI_Widget_Enable(TSI_WidgetPointers[i]);
    }
}

void TSI_Widget_DisableAll(void)
{
    int i;
    for (i = 0; i < TSI_TOTAL_WIDGETS_NUM; i++)
    {
        TSI_Widget_Disable(TSI_WidgetPointers[i]);
    }
}

void TSI_Widget_ApplyConfig(TSI_WidgetTypeDef* widget)
{
    int i;
    TSI_SensorTypeDef* sensor = widget->meta->sensors;
    
    /* Apply all sensors' configurations */
    for (i = 0; i < widget->meta->sensorCnt; i++)
    {
        TSI_Sensor_ApplyConfig(sensor);
        sensor++;
    }
}

void TSI_Widget_Enable(TSI_WidgetTypeDef* widget)
{
    int i;
    TSI_SensorTypeDef* sensor = widget->meta->sensors;
    
    /* Set widget enable status */
    widget->enable = TSI_WIDGET_ENABLE;
    
    /* Enable all sensors */
    for (i = 0; i < widget->meta->sensorCnt; i++)
    {
        TSI_Sensor_Enable(sensor);
        sensor++;
    }
}

void TSI_Widget_Disable(TSI_WidgetTypeDef* widget)
{
    int i;
    TSI_SensorTypeDef* sensor = widget->meta->sensors;
    
    /* Set widget enable status */
    widget->enable = TSI_WIDGET_DISABLE;
    
    /* Disable all sensors */
    for (i = 0; i < widget->meta->sensorCnt; i++)
    {
        TSI_Sensor_Disable(sensor);
        sensor++;
    }
}

void TSI_Widget_UpdateAll(void)
{
    int i;
    for (i = 0; i < TSI_TOTAL_WIDGETS_NUM; i++)
    {
        TSI_Widget_Update(TSI_WidgetPointers[i]);
    }
}

void TSI_Widget_Update(TSI_WidgetTypeDef* widget)
{
    int i;
    TSI_SensorTypeDef* sensor = widget->meta->sensors;
    uint8_t sensorActive = 0U;
    
    /* Update all sensors */
    for (i = 0; i < widget->meta->sensorCnt; i++)
    {
        TSI_Sensor_Update(sensor);
        if((sensor->status & 0x1U) == 0x1U) 
        {
            /* If any sensor is activated, mark widget general status as active */
            sensorActive = 1U;
        }
        sensor++;
    }

    /* Update widget general status */
    widget->status = sensorActive;

    /* Update widget status */
    switch (widget->meta->type)
    {
    case TSI_WIDGET_BUTTON:
        TSI_Widget_UpdateButton((TSI_ButtonWidgetTypeDef*)widget);
        break;
    
    case TSI_WIDGET_PROXIMITY:
        TSI_Widget_UpdateProximity((TSI_ProximityWidgetTypeDef*)widget);
        break;

    case TSI_WIDGET_SLIDER:
        TSI_Widget_UpdateSlider((TSI_SliderWidgetTypeDef*)widget);
        break;
    }
}

void TSI_Widget_UpdateButton(TSI_ButtonWidgetTypeDef* buttonWidget)
{
    /* Button widget is activated when any sensor is activated */
    buttonWidget->buttonStatus = buttonWidget->base.status;
}

void TSI_Widget_UpdateProximity(TSI_ProximityWidgetTypeDef* proximityWidget)
{
    if(proximityWidget->base.status == 1U)
    {
        /* Proximity widget is activated when any sensor is activated */
        proximityWidget->proximityStatus = 2U;
    }
    else 
    {
        int i;
        const TSI_MetaWidgetTypeDef *metaWidget = proximityWidget->base.meta;
        TSI_SensorTypeDef* sensor = metaWidget->sensors;

        /* Check if any sensor is in proximity status */
        for (i = 0; i < metaWidget->sensorCnt; i++)
        {
            if ((sensor->status & 0x2U) == 0x2U)
            {
                /* Proximity widget is in proximity status when any sensor is in proximity status */
                proximityWidget->proximityStatus = 1U;
                return;
            }
            sensor++;
        }

        proximityWidget->proximityStatus = 0U;
    }
}

void TSI_Widget_UpdateSlider(TSI_SliderWidgetTypeDef* sliderWidget)
{
    if(sliderWidget->base.status == 1U)
    {
        int i;
        uint8_t peakIndex = 0U;
        uint16_t peakSignalVal = 0U;
        uint16_t centroidCalcData[3];
        uint16_t threshold = sliderWidget->base.activeTh - sliderWidget->base.activeHys;
        int32_t centroid;
        const TSI_MetaWidgetTypeDef *metaWidget = sliderWidget->base.meta;
        TSI_SensorTypeDef* sensor = metaWidget->sensors;
        int32_t mul = (TSI_SLIDER_RESOLUTION / (metaWidget->sensorCnt - 1)) << 8U;

        /* 
            Calculate the slider postion using centroid algorithm:
            1. Find the sensor X with the highest signal, which is the center of the touched area;
            2. Calcuate centroid using X's neighbour sensor;
        */
        
        /* Find peak value and corresponding sensor */
        for (i = 0; i < metaWidget->sensorCnt; i++)
        {
            if(sensor->diffCount > threshold && sensor->diffCount > peakSignalVal) 
            {
                peakIndex = i+1;
                peakSignalVal = sensor->diffCount;
            }
            sensor++;
        }
        
        if(peakIndex != 0)
        {
            /* Get neighbour sensors signal value */
            sensor = metaWidget->sensors;
            if(peakIndex == 1)
            {
                centroidCalcData[0] = 0;
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = (int32_t)sensor[1].diffCount;
            }
            else if(peakIndex == metaWidget->sensorCnt)
            {
                centroidCalcData[0] = sensor[peakIndex-2].diffCount;
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = 0;
            }
            else
            {
                centroidCalcData[0] = sensor[peakIndex-2].diffCount;
                centroidCalcData[1] = peakSignalVal;
                centroidCalcData[2] = sensor[peakIndex].diffCount;
            }
            
            /* Calculate centroid */
            centroid = ((int32_t)centroidCalcData[2] 
                        - (int32_t)centroidCalcData[0]) * mul / 
                        ((int32_t)centroidCalcData[0]
                        + (int32_t)centroidCalcData[1]
                        + (int32_t)centroidCalcData[2]);
            centroid += (peakIndex-1) * mul;
            centroid >>= 8U;
            sliderWidget->centerPos = centroid;
        }
    }
}

/********************** (C) COPYRIGHT Fudan Microelectronics **** END OF FILE ***********************/
