#include "tsi_debug.h"
#include "tsi_debug_io.h"
#include "tsi.h"
#include <string.h>

#define TSI_DBG_HEADER_ID                       0x25U
#define TSI_DBG_BULK_HEADER_ID                  0x24U

#define TSI_DBG_STAT_RST                        0
#define TSI_DBG_STAT_HEADER_WAIT                1
#define TSI_DBG_STAT_HEADER_WAIT_ACK_CPLT       2
#define TSI_DBG_STAT_SEND_WAIT_ACK              3
#define TSI_DBG_STAT_RECV_WAIT_DATA             4
#define TSI_DBG_STAT_RECV_WAIT_ACK_CPLT         5
#define TSI_DBG_STAT_HNDSHK_WAIT                6
#define TSI_DBG_STAT_HNDSHK_WAIT_ACK_CPLT       7
#define TSI_DBG_STAT_BULK_READ_WAIT_CPLT        8

static const uint8_t CRC4_TABLE[256] = {
	0x0, 0x7, 0xE, 0x9, 0x5, 0x2, 0xB, 0xC, 0xA, 0xD, 0x4, 0x3, 0xF, 0x8, 0x1, 0x6,
	0xD, 0xA, 0x3, 0x4, 0x8, 0xF, 0x6, 0x1, 0x7, 0x0, 0x9, 0xE, 0x2, 0x5, 0xC, 0xB,
	0x3, 0x4, 0xD, 0xA, 0x6, 0x1, 0x8, 0xF, 0x9, 0xE, 0x7, 0x0, 0xC, 0xB, 0x2, 0x5,
	0xE, 0x9, 0x0, 0x7, 0xB, 0xC, 0x5, 0x2, 0x4, 0x3, 0xA, 0xD, 0x1, 0x6, 0xF, 0x8,
	0x6, 0x1, 0x8, 0xF, 0x3, 0x4, 0xD, 0xA, 0xC, 0xB, 0x2, 0x5, 0x9, 0xE, 0x7, 0x0,
	0xB, 0xC, 0x5, 0x2, 0xE, 0x9, 0x0, 0x7, 0x1, 0x6, 0xF, 0x8, 0x4, 0x3, 0xA, 0xD,
	0x5, 0x2, 0xB, 0xC, 0x0, 0x7, 0xE, 0x9, 0xF, 0x8, 0x1, 0x6, 0xA, 0xD, 0x4, 0x3,
	0x8, 0xF, 0x6, 0x1, 0xD, 0xA, 0x3, 0x4, 0x2, 0x5, 0xC, 0xB, 0x7, 0x0, 0x9, 0xE,
	0xC, 0xB, 0x2, 0x5, 0x9, 0xE, 0x7, 0x0, 0x6, 0x1, 0x8, 0xF, 0x3, 0x4, 0xD, 0xA,
	0x1, 0x6, 0xF, 0x8, 0x4, 0x3, 0xA, 0xD, 0xB, 0xC, 0x5, 0x2, 0xE, 0x9, 0x0, 0x7,
	0xF, 0x8, 0x1, 0x6, 0xA, 0xD, 0x4, 0x3, 0x5, 0x2, 0xB, 0xC, 0x0, 0x7, 0xE, 0x9,
	0x2, 0x5, 0xC, 0xB, 0x7, 0x0, 0x9, 0xE, 0x8, 0xF, 0x6, 0x1, 0xD, 0xA, 0x3, 0x4,
	0xA, 0xD, 0x4, 0x3, 0xF, 0x8, 0x1, 0x6, 0x0, 0x7, 0xE, 0x9, 0x5, 0x2, 0xB, 0xC,
	0x7, 0x0, 0x9, 0xE, 0x2, 0x5, 0xC, 0xB, 0xD, 0xA, 0x3, 0x4, 0x8, 0xF, 0x6, 0x1,
	0x9, 0xE, 0x7, 0x0, 0xC, 0xB, 0x2, 0x5, 0x3, 0x4, 0xD, 0xA, 0x6, 0x1, 0x8, 0xF,
	0x4, 0x3, 0xA, 0xD, 0x1, 0x6, 0xF, 0x8, 0xE, 0x9, 0x0, 0x7, 0xB, 0xC, 0x5, 0x2
};

/**
 *  @brief Debug protocol status
 */
static uint8_t state;

/**
 *  @brief Debug protocol handshake sync buffer
 */
static uint8_t protoSync;

/**
 *  @brief Debug protocol header buffer
 */
static uint8_t protoHeader[4];

/**
 *  @brief Debug protocol ack buffer
 */
static uint8_t protoAck;

/**
 *  @brief Debug protocol trace buffer and its shadow
 */
static uint8_t protoTraceBuffer[TRACE_BUFFER_MAX_SIZE];
static uint8_t protoTraceBufferShadow[TRACE_BUFFER_MAX_SIZE];
static int protoTraceBufferSize = 0;
static uint8_t protoTraceBufferUpdateFlag = 0;

/**
 *  @brief Debug protocol trace vars
 */
static TSI_Debug_TraceVarTypeDef traceVars[TRACE_VAR_MAX_NUM];

/**
 *  @brief Calculate buffer size
 */
static uint16_t CalcBufferSize(void)
{
    int index, size;
    for (index = 0, size = 0; index < TRACE_VAR_MAX_NUM; index++)
    {
        if(traceVars[index].valid)
        {
            size += traceVars[index].size + 1;
        }  
    }

    if(size != 0U)
    {        
        // CRC-4 at the end
        size += 1;
    }
        
    return size;
}

/**
 *  @brief Set new trace var
 */
static int AddTraceVar(uint16_t addr, uint8_t size)
{
    int i;
    for(i = 0; i < TRACE_VAR_MAX_NUM; i++)
    {
        // Find empty slot
        if(!traceVars[i].valid)
        {
            traceVars[i].valid = 1;
            traceVars[i].addr = addr;
            traceVars[i].size = size;
            
            protoTraceBufferSize = CalcBufferSize();
            
            // Ignore buffer update before operation
            if(protoTraceBufferUpdateFlag == 1U)
            {
                protoTraceBufferUpdateFlag = 0U;
            }
            
            return i;
        }
    }
    return -1;
}

/**
 *  @brief Remove trace variable
 */
static void RemoveTraceVar(uint8_t index)
{
    traceVars[index].valid = 0;
    protoTraceBufferSize = CalcBufferSize();
    
    // Ignore buffer update before operation
    if(protoTraceBufferUpdateFlag == 1U)
    {
        protoTraceBufferUpdateFlag = 0U;
    }
}

/**
 *  @brief Remove all trace variables
 */
static void RemoveAllTraceVars(void)
{
    int i;
    for(i = 0; i < TRACE_VAR_MAX_NUM; i++)
    {
        if(traceVars[i].valid)
        {
            traceVars[i].valid = 0;
        }
    }
    
    protoTraceBufferSize = CalcBufferSize();
    
    // Ignore buffer update before operation
    if(protoTraceBufferUpdateFlag == 1U)
    {
        protoTraceBufferUpdateFlag = 0U;
    }
}

/**
 *  @brief Reset trace buffer
 */
static void ResetTraceBuffer(void)
{   
    // Clear buffer update flag and reset shadow buffer
    protoTraceBufferUpdateFlag = 0U;
    memset(protoTraceBuffer, 0U, sizeof(protoTraceBuffer));
    memset(protoTraceBufferShadow, 0U, sizeof(protoTraceBufferShadow));
}

/**
 *  @brief Data buffer for the transfer
 *  @warning Debug software should take care of the endian!
 */
static uint8_t dataBuffer[sizeof(uint32_t) * 4 + 1];

// Debug IO signal slot
static void TSI_DebugIO_signalSlot(TSI_DebugIOSignal sig)
{
    if(sig == TSI_DebugIO_SigTxComplete)
    {
        /* Header ack sent */
        if(state == TSI_DBG_STAT_HEADER_WAIT_ACK_CPLT)
        {
            // Flush buffer
            memset(&protoAck, 0, sizeof(protoAck));
            
            // If it's a read access, send the data to the master, and prepare for receiving ack
            if(HEADER_DIR(protoHeader) == 0)
            {
                // Get the total data size in bytes from header
                uint32_t size = (1 << (HEADER_ACCESS_WIDTH(protoHeader) - 1)) * 
                                    (HEADER_ACCESS_SIZE(protoHeader) + 1);
                
                // Note: Protocol should do nothing when this transmit completed
                TSI_DebugIO.transmit(dataBuffer, size + 1); 
                TSI_DebugIO.receive(&protoAck, sizeof(protoAck));
                
                // Indicate that the debug protocol is waiting for the master's ack
                state = TSI_DBG_STAT_SEND_WAIT_ACK;
            }
            // If it's a write access, change the debug status
            else
            {
                // Indicate that the debug protocol is waiting for the master's data
                state = TSI_DBG_STAT_RECV_WAIT_DATA;
            }
        }
        /* Ack sent in receive mode (master send) */
        else if (state == TSI_DBG_STAT_RECV_WAIT_ACK_CPLT)
        {
            // flush buffer
            memset(protoHeader, 0, sizeof(protoHeader));
            memset(&protoAck, 0, sizeof(protoAck));
            memset(dataBuffer, 0, sizeof(dataBuffer));
            
            // Prepare to receive the header
            TSI_DebugIO.receive(protoHeader, sizeof(protoHeader));
            
            // Indicate that the debug protocol is waiting for the header
            state = TSI_DBG_STAT_HEADER_WAIT;
        }
        /* Data sent in bulk-read mode */
        else if (state == TSI_DBG_STAT_BULK_READ_WAIT_CPLT)
        {
            // flush buffer
            memset(protoHeader, 0, sizeof(protoHeader));
             
            // Prepare to receive the header
            TSI_DebugIO.receive(protoHeader, sizeof(protoHeader));
            
            // Indicate that the debug protocol is waiting for the header
            state = TSI_DBG_STAT_HEADER_WAIT;
        }
        /* Ack sent in handshake mode */
        else if (state == TSI_DBG_STAT_HNDSHK_WAIT_ACK_CPLT)
        {
            // Restart protocol
            TSI_Debug_Restart();
        }
    }
    else if(sig == TSI_DebugIO_SigRxComplete)
    {
        /* Header received */
        if(state == TSI_DBG_STAT_HEADER_WAIT)
        {
            TSI_Debug_ErrorCode err = TSI_DBG_ERR_NONE;
            
            /**
             *  Handle Frame Header
             */
            if(((HEADER_ID(protoHeader) == TSI_DBG_HEADER_ID) &&
                    (HEADER_ACCESS_WIDTH(protoHeader) != 0x00)) ||
                    (HEADER_ID(protoHeader) == TSI_DBG_BULK_HEADER_ID))
            {
                /* Check the CRC-4 */
                int i;
                uint8_t tmp = 0x00;
                uint32_t target = (((uint32_t)protoHeader[3] << 24) +
                                   ((uint32_t)protoHeader[2] << 16) +
                                   ((uint32_t)protoHeader[1] << 8) +
                                   ((uint32_t)protoHeader[0]));
                
                // Fill CRC result bits with 1 while calculating
                target |= (0xF << 8);
                
                // Calculate CRC-4
                for(i=0; i<4; i++)
                {
                    tmp = CRC4_TABLE[tmp ^ (target & 0xFFU)];
                    target >>= 8;
                }
                
                // Parity error
                if(tmp != HEADER_PARITY(protoHeader)) err = TSI_DBG_ERR_HEAD_PARITY;                
            }
            else if((protoHeader[0] == 0xFF) &&
                        (protoHeader[1] == 0xFF) &&
                        (protoHeader[2] == 0xFF) &&
                        (protoHeader[3] == 0xFF))
            {
                /** 
                 *  If the master cannot recover from error state(error for 
                 *  too many times or timeout when waiting ack or data), 
                 *  master should send this handshake signal(header-sized, 
                 *  all 0xFF) continuously until it receives RHS ack from the slave.
                 *  It's fast when the slave is waiting header, and slow when the
                 *  slave is waiting for master's data.
                 *  
                 *  After receiving the handshake signal, the slave should return
                 *  an RHS ack and prepare to receive at the same time. The 
                 *  master then start to transmit a single-byte sync signal(0xEA) 
                 *  with a 50ms interval. Slave should return an EHS signal and 
                 *  restart the protocol as soon as possible when it received 
                 *  the sync signal. Master must wait for at least 10ms after receiving
                 *  EHS before the protocol restarting.
                 */            
                // Send RHS ack back
                WRITE_ACK_STATUS(protoAck, TSI_DBG_RHS);
                WRITE_ACK_ERRCODE(protoAck, TSI_DBG_ERR_NONE);
                TSI_DebugIO.transmit(&protoAck, sizeof(protoAck));
                
                // Ready to receive sync signal
                TSI_DebugIO.receive(&protoSync, 1);
                
                // Indicate that the debug protocol is waiting for the sync signal
                state = TSI_DBG_STAT_HNDSHK_WAIT;
                
                return;
            }
            else
            {
                // Format error
                err = TSI_DBG_ERR_HEAD_FORMAT;
            }

            if(err != TSI_DBG_ERR_NONE)
            {
                /**
                 *  If it's not the header or the header contains error, then 
                 *  flush all buffer and try again
                 */
                // Flush all buffers
                memset(protoHeader, 0, sizeof(protoHeader));
                TSI_DebugIO.clear();
                
                // Receive again
                TSI_DebugIO.receive(protoHeader, sizeof(protoHeader));
            }
            else
            {
                /**
                 *  If the header is ok, prepare for the transfer and send 
                 *  RDY ack back
                 */  
                // Flush receive buffers
                TSI_DebugIO.clear();
                
                if(HEADER_ID(protoHeader) == TSI_DBG_HEADER_ID)
                {
                    // Prepare for the transfer
                    if(HEADER_DIR(protoHeader) == 0)
                    {
                        int i;
                        uint8_t tmp = 0x00;
                        
                        /* Read access, prepare the data first */
                        // Point the base pointer to the TSI_Module object 
                        void *pBase = &TSI_Module;
                        
                        // Get the offset from header
                        uint16_t offset = HEADER_OFFSET(protoHeader);
                        
                        // Get the total data size in bytes from header
                        uint32_t size = (1 << (HEADER_ACCESS_WIDTH(protoHeader) - 1)) * 
                                            (HEADER_ACCESS_SIZE(protoHeader) + 1);
                        
                        // Get the target data
                        memcpy((uint8_t*)dataBuffer, ((uint8_t*)pBase + offset), size);
                        
                        // Calculate the CRC-4
                        for(i=0; i<size; i++)
                        {
                            tmp = CRC4_TABLE[tmp ^ dataBuffer[i]];
                        }
                        
                        // Write the CRC-4 to the end
                        dataBuffer[i] = tmp;
                    }
                    else
                    {
                        // Get the total data size in bytes from header
                        uint32_t size = (1 << (HEADER_ACCESS_WIDTH(protoHeader) - 1)) * 
                                            (HEADER_ACCESS_SIZE(protoHeader) + 1);
                        
                        /* Write access, prepare to receive master's data */
                        // Flush buffer
                        memset(dataBuffer, 0, sizeof(dataBuffer));
                        
                        // Start to receive
                        TSI_DebugIO.receive(dataBuffer, size + 1);
                    }
                
                    // Send RDY ack back
                    WRITE_ACK_STATUS(protoAck, TSI_DBG_RDY);
                    WRITE_ACK_ERRCODE(protoAck, TSI_DBG_ERR_NONE);
                    TSI_DebugIO.transmit(&protoAck, sizeof(protoAck));
                    
                    // Indicate that the debug protocol is waiting for the header ack transfer to complete
                    state = TSI_DBG_STAT_HEADER_WAIT_ACK_CPLT;
                }
                else if(HEADER_ID(protoHeader) == TSI_DBG_BULK_HEADER_ID)
                {
                    // Check if there's update
                    if(protoTraceBufferUpdateFlag != 0U)
                    {
                        protoTraceBufferUpdateFlag = 0U;
                        
                        if(protoTraceBufferSize > 0U)
                        {
                            memcpy(protoTraceBuffer, protoTraceBufferShadow, 
                                protoTraceBufferSize);
                        }
                    }
                    
                    if(protoTraceBufferSize > 0U)
                    {
                        // Start transfer
                        TSI_DebugIO.transmit(protoTraceBuffer, protoTraceBufferSize);

                        // Indicate that the debug protocol is waiting for the bulk-read transfer to complete
                        state = TSI_DBG_STAT_BULK_READ_WAIT_CPLT;
                    }
                    else
                    {
                        // No data to transfer 
                        // flush buffer
                        memset(protoHeader, 0, sizeof(protoHeader));

                        // Prepare to receive the header
                        TSI_DebugIO.receive(protoHeader, sizeof(protoHeader));

                        // Indicate that the debug protocol is waiting for the header
                        state = TSI_DBG_STAT_HEADER_WAIT;
                    }
                }
            }
        }
        /* Master ack received in send mode (master receive) */
        else if (state == TSI_DBG_STAT_SEND_WAIT_ACK)
        {
            /** ---------------------------------------------------------------*
             *  Currently there is no retry mechanism implemented, so the slave
             *  just ignore the ack's content.
             *
             *  Further we may implement it here if necessary. 
             * ---------------------------------------------------------------*/
            
            // flush buffer
            memset(protoHeader, 0, sizeof(protoHeader));
            memset(&protoAck, 0, sizeof(protoAck));
            memset(dataBuffer, 0, sizeof(dataBuffer));
            TSI_DebugIO.clear();
            
            // Prepare to receive the header
            TSI_DebugIO.receive(protoHeader, sizeof(protoHeader));
            
            // Indicate that the debug protocol is waiting for the header
            state = TSI_DBG_STAT_HEADER_WAIT;
        }
        /* Master data received in receive mode (master send) */
        else if (state == TSI_DBG_STAT_RECV_WAIT_DATA)
        {
            int i;
            uint8_t tmp = 0x00;
            
            // Point the base pointer to the TSI_Module object 
            void *pBase = &TSI_Module;
            
            // Get the offset from header
            uint16_t offset = HEADER_OFFSET(protoHeader);
            
            // Get the total data size in bytes from header
            uint32_t size = (1 << (HEADER_ACCESS_WIDTH(protoHeader) - 1)) * 
                                (HEADER_ACCESS_SIZE(protoHeader) + 1);
            
            // Calculate the CRC-4
            for(i=0; i<size; i++)
            {
                tmp = CRC4_TABLE[tmp ^ dataBuffer[i]];
            }
            
            // CRC-4 mismatched
            if(dataBuffer[i] != tmp) 
            {
                /** ---------------------------------------------------------------*
                 *  Currently there is no retry mechanism implemented, and master
                 *  will start a new transmission when received ERR ack. 
                 *  
                 *  Further we may implement it here if necessary.
                 * ---------------------------------------------------------------*/
                
                // Send ERR ack back
                WRITE_ACK_STATUS(protoAck, TSI_DBG_ERR);
                WRITE_ACK_ERRCODE(protoAck, TSI_DBG_ERR_DATA_PARITY);
                TSI_DebugIO.transmit(&protoAck, sizeof(protoAck));
            }
            // CRC-4 matched
            else
            {
                // Set the data
                memcpy(((uint8_t*)pBase + offset), (uint8_t*)dataBuffer, size);

                // Send EOT ack back
                WRITE_ACK_STATUS(protoAck, TSI_DBG_EOT);
                WRITE_ACK_ERRCODE(protoAck, TSI_DBG_ERR_NONE);
                TSI_DebugIO.transmit(&protoAck, sizeof(protoAck));
            }
            
            // Indicate that the debug protocol is waiting for the receive ack transfer to complete
            state = TSI_DBG_STAT_RECV_WAIT_ACK_CPLT;
        }
        /* Sync signal received */
        else if (state == TSI_DBG_STAT_HNDSHK_WAIT)
        {
            if(protoSync == 0xEA)
            {
                // Send EHS ack back
                WRITE_ACK_STATUS(protoAck, TSI_DBG_EHS);
                WRITE_ACK_ERRCODE(protoAck, TSI_DBG_ERR_NONE);
                TSI_DebugIO.transmit(&protoAck, sizeof(protoAck));
                
                // Indicate that the debug protocol is waiting for the sync ack transfer to complete
                state = TSI_DBG_STAT_HNDSHK_WAIT_ACK_CPLT;
            }
            else
            {
                // Not the sync signal, receive again
                TSI_DebugIO.receive(&protoSync, 1);
            }   
        }
    }
}

// Init debug protocol
void TSI_Debug_Init(void)
{
    // Register the debug IO signal slot
    TSI_DebugIO.emitSignal = TSI_DebugIO_signalSlot;
    
    // Init debug interface
    TSI_DebugIO.init();
    
    // Init debug protocol buffer
    memset(protoHeader, 0, sizeof(protoHeader));
    memset(&protoAck, 0, sizeof(protoAck));
    memset(dataBuffer, 0, sizeof(dataBuffer));
    TSI_DebugIO.clear();
    
    // Init debug state
    state = TSI_DBG_STAT_RST;
}

// Start debug protocol
void TSI_Debug_Start(void)
{
    // Prepare to receive the header
    TSI_DebugIO.receive(protoHeader, sizeof(protoHeader));
    
    // Indicate that the debug protocol is waiting for the header
    state = TSI_DBG_STAT_HEADER_WAIT;
}

// Restart debug protocol
void TSI_Debug_Restart(void)
{
    // Re-Init debug protocol buffer
    memset(protoHeader, 0, sizeof(protoHeader));
    memset(&protoAck, 0, sizeof(protoAck));
    memset(dataBuffer, 0, sizeof(dataBuffer));
    TSI_DebugIO.clear();
    
    // Restart the debug protocol
    TSI_Debug_Start();
}

// Stop debug protocol
void TSI_Debug_Stop(void)
{    
    // Init debug state
    state = TSI_DBG_STAT_RST;
    
    // Init debug protocol buffer
    memset(protoHeader, 0, sizeof(protoHeader));
    memset(&protoAck, 0, sizeof(protoAck));
    memset(dataBuffer, 0, sizeof(dataBuffer));
    TSI_DebugIO.clear();
}

// Deinit debug protocol
void TSI_Debug_Deinit(void)
{
    TSI_DebugIO.deinit();
}

/* Command Handler -----------------------------------------------------------*/
#define TSI_DEBUG_CMD_TYPE_CTRL         0x01
#define TSI_DEBUG_CMD_TYPE_ADD_TRACE    0x02 
#define TSI_DEBUG_CMD_TYPE_RMV_TRACE    0x03 
#define TSI_DEBUG_CMD_TYPE_CLR_TRACE    0x04
#define TSI_DEBUG_CMD_TYPE_QUERY_STAT   0xFE

#define TSI_DEBUG_CMD_OP1_START         0x01
#define TSI_DEBUG_CMD_OP1_STOP          0x02
#define TSI_DEBUG_CMD_OP1_RECFG         0x03

#define TSI_DEBUG_COMMAND_TYPE(__CMD__)  ((__CMD__ & (0xFFU << 24)) >> 24)
#define TSI_DEBUG_COMMAND_OP1(__CMD__)  ((__CMD__ & (0xFFFFU << 8)) >> 8)
#define TSI_DEBUG_COMMAND_OP2(__CMD__)  (__CMD__ & 0xFFU)
#define TSI_DEBUG_COMMAND_STATUS(__OP1__, __OP2__)  \
    (((uint32_t)TSI_DEBUG_CMD_TYPE_QUERY_STAT << 24) |  \
     ((uint32_t)__OP1__ << 8) | \
     ((uint32_t)__OP2__))

// Handle debug protocol event
void TSI_Debug_Handler(void)
{
    // Handle debug command
    if(TSI_DEBUG_COMMAND_TYPE(TSI_Module.debugCmd) == TSI_DEBUG_CMD_TYPE_CTRL)
    {
        
        if(TSI_DEBUG_COMMAND_OP1(TSI_Module.debugCmd) == TSI_DEBUG_CMD_OP1_START)
        {
            // Start(Resume)
            TSI_Resume();
            
            // Set command status to OK
            TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(0, 0);
        }
        else if(TSI_DEBUG_COMMAND_OP1(TSI_Module.debugCmd) == TSI_DEBUG_CMD_OP1_STOP)
        {
            // Stop(Suspend)
            TSI_Suspend();
            
            // Set command status to OK
            TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(0, 0);
        }
        else if(TSI_DEBUG_COMMAND_OP1(TSI_Module.debugCmd) == TSI_DEBUG_CMD_OP1_RECFG)
        {
            // Apply configuration
            TSI_Reconfig();
            
            // Set command status to OK
            TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(0, 0);
        }
    }
    else if (TSI_DEBUG_COMMAND_TYPE(TSI_Module.debugCmd) == TSI_DEBUG_CMD_TYPE_ADD_TRACE)
    {
        // Add trace var at OP1, size OP2
        int index = AddTraceVar(TSI_DEBUG_COMMAND_OP1(TSI_Module.debugCmd), TSI_DEBUG_COMMAND_OP2(TSI_Module.debugCmd));
        
        // Reset Trace Buffer
        ResetTraceBuffer();

        if(index >= 0)
        {
            // Set command status to OK
            TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(protoTraceBufferSize, index);
        }
        else
        {
            // Set command status to NG
            TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(0, 2);
        }
    }
    else if (TSI_DEBUG_COMMAND_TYPE(TSI_Module.debugCmd) == TSI_DEBUG_CMD_TYPE_RMV_TRACE)
    {
        if(TSI_DEBUG_COMMAND_OP1(TSI_Module.debugCmd) == 0)
        {
            // Remove trace var at index OP1
            RemoveTraceVar(TSI_DEBUG_COMMAND_OP2(TSI_Module.debugCmd));

            // Set command status to OK
            TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(protoTraceBufferSize, 0);
        }
        else 
        {
            // Command frame error
            TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(0, 1);
        }
    }
    else if (TSI_DEBUG_COMMAND_TYPE(TSI_Module.debugCmd) == TSI_DEBUG_CMD_TYPE_CLR_TRACE)
    {
        // Set command status to OK
        TSI_Module.debugCmd = TSI_DEBUG_COMMAND_STATUS(0, 0);
        
        // Remove all
        RemoveAllTraceVars();
    }
}

void TSI_Debug_UpdateTrace(void)
{
    int i;
    uint8_t crc = 0x00;
    void *pBase = &TSI_Module;
    uint8_t* pBuffer = protoTraceBufferShadow;
    
    if((protoTraceBufferUpdateFlag == 0U) && (protoTraceBufferSize > 0U))
    {
        // Update buffer
        for (i = 0; i < TRACE_VAR_MAX_NUM; i++)
        {
            if(traceVars[i].valid)
            {
                // record var index before var contents
                *pBuffer++ = (uint8_t)i;

                // record var contents
                memcpy(pBuffer, ((uint8_t*)pBase + traceVars[i].addr), traceVars[i].size);
                
                // increase pointer
                pBuffer += traceVars[i].size;
            }
        }

        // Calculate CRC-4
        for(i = 0; i < protoTraceBufferSize - 1; i++)
        {
            crc = CRC4_TABLE[crc ^ protoTraceBufferShadow[i]];
        }
        *pBuffer = crc;
        
        // Notify update
        protoTraceBufferUpdateFlag = 1U;
    }
}
